package net.knarcraft.dynmapcitizens.handler.trait.quests;

import me.blackvein.quests.quests.IQuest;
import net.knarcraft.dynmapcitizens.property.Icon;

import java.util.ArrayList;
import java.util.List;

/**
 * Keeps track of quest information for an NPC
 */
public class NPCQuestInfo {

    private final List<IQuest> questStart = new ArrayList<>();
    private final List<IQuest> questKill = new ArrayList<>();
    private final List<IQuest> questDeliver = new ArrayList<>();
    private final List<IQuest> questInteract = new ArrayList<>();

    /**
     * Adds the given quest to the list of quests this NPC provides
     *
     * @param quest <p>The quest to add</p>
     */
    public void addQuestStart(IQuest quest) {
        this.questStart.add(quest);
    }

    /**
     * Adds the given quest to the list of quests in which this NPC is killed
     *
     * @param quest <p>The quest to add</p>
     */
    public void addQuestKill(IQuest quest) {
        this.questKill.add(quest);
    }

    /**
     * Adds the given quest to the list of quests in which this NPC takes deliveries
     *
     * @param quest <p>The quest to add</p>
     */
    public void addQuestDeliver(IQuest quest) {
        this.questDeliver.add(quest);
    }

    /**
     * Adds the given quest to the list of quests in which this NPC has to be interacted with
     *
     * @param quest <p>The quest to add</p>
     */
    public void addQuestInteract(IQuest quest) {
        this.questInteract.add(quest);
    }

    /**
     * Gets all quests given by this NPC
     *
     * @return <p>All quests this NPC is used to start</p>
     */
    public List<IQuest> getQuestStarts() {
        return new ArrayList<>(this.questStart);
    }

    /**
     * Gets all quests in which this NPC has to be killed
     *
     * @return <p>All kill quests involving this NPC</p>
     */
    public List<IQuest> getQuestKills() {
        return new ArrayList<>(this.questKill);
    }

    /**
     * Gets all quests requiring something be delivered to this NPC
     *
     * @return <p>All quests delivering to this NPC</p>
     */
    public List<IQuest> getQuestDeliveries() {
        return new ArrayList<>(this.questDeliver);
    }

    /**
     * Gets all quests requiring interaction with this NPC
     *
     * @return <p>All quests requiring interaction with this NPC</p>
     */
    public List<IQuest> getQuestInteractions() {
        return new ArrayList<>(this.questInteract);
    }

    /**
     * Gets the main type of the quest the NPC
     *
     * @return <p>The main type of the quest NPC</p>
     */
    public QuestNPCType getQuestNPCType() {
        if (!questStart.isEmpty()) {
            return QuestNPCType.GIVER;
        } else if (!questKill.isEmpty() && questInteract.isEmpty() && questDeliver.isEmpty()) {
            return QuestNPCType.KILL;
        } else if (!questInteract.isEmpty() && questKill.isEmpty() && questDeliver.isEmpty()) {
            return QuestNPCType.INTERACT;
        } else if (!questDeliver.isEmpty() && questKill.isEmpty() && questInteract.isEmpty()) {
            return QuestNPCType.DELIVER;
        } else {
            return QuestNPCType.CHAIN;
        }
    }

    /**
     * Gets the Icon to use when marking this NPC
     *
     * @return <p>The icon used to mark this NPC</p>
     */
    public Icon getNPCIcon() {
        QuestNPCType type = getQuestNPCType();
        return switch (type) {
            case KILL -> Icon.QUEST_KILL;
            case GIVER -> Icon.QUEST_GIVER;
            case DELIVER -> Icon.QUEST_DELIVER;
            case INTERACT -> Icon.QUEST_INTERACT;
            case CHAIN -> Icon.QUEST_CHAIN;
        };
    }

}
