package net.knarcraft.dynmapcitizens.handler.trait.quests;

import me.blackvein.quests.QuestsAPI;
import me.blackvein.quests.quests.IQuest;
import me.blackvein.quests.quests.IStage;
import net.citizensnpcs.api.CitizensAPI;
import net.citizensnpcs.api.npc.NPCRegistry;
import net.knarcraft.dynmapcitizens.DynmapCitizens;
import net.knarcraft.dynmapcitizens.handler.trait.AbstractTraitHandler;
import net.knarcraft.dynmapcitizens.property.Icon;
import net.knarcraft.dynmapcitizens.util.QuestsHelper;
import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.World;
import org.dynmap.DynmapAPI;
import org.dynmap.markers.CircleMarker;
import org.dynmap.markers.GenericMarker;
import org.dynmap.markers.MarkerIcon;
import org.dynmap.markers.MarkerSet;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.logging.Level;

/**
 * A handler class for the quests trait
 */
public class QuestsHandler extends AbstractTraitHandler {

    private QuestsAPI questsAPI;
    private MarkerSet questMarkerSet;
    private MarkerSet questAreaMarkerSet;
    private Map<Icon, MarkerIcon> markerIcons;
    private Collection<IQuest> loadedQuests;
    private Map<UUID, NPCQuestInfo> questGiverInfo;

    @Override
    public void initialize() {
        questsAPI = (QuestsAPI) Bukkit.getServer().getPluginManager().getPlugin("Quests");
        DynmapAPI dynmapAPI = DynmapCitizens.getInstance().getDynmapAPI();
        markerIcons = DynmapCitizens.getInstance().getMarkerIcons();
        if (questsAPI != null) {
            questMarkerSet = getMarkerSet(dynmapAPI, "quests", "Quests");
            questAreaMarkerSet = getMarkerSet(dynmapAPI, "quest_areas", "Quest areas");
            if (questMarkerSet != null && questAreaMarkerSet != null) {
                questMarkerSet.setHideByDefault(false);
                questAreaMarkerSet.setHideByDefault(true);
                questMarkerSet.setLayerPriority(3);
                questAreaMarkerSet.setLayerPriority(2);
                isEnabled = true;
                return;
            }
        }
        isEnabled = false;
    }

    @Override
    public void updateMarkers() {
        if (questsAPI.isLoading()) {
            return;
        }

        //There is no point in updating if there has been no changes in quests
        boolean questsChanged = loadedQuests == null || !loadedQuests.equals(questsAPI.getLoadedQuests());
        loadedQuests = questsAPI.getLoadedQuests();

        //Remove old quest markers
        questMarkerSet.getMarkers().forEach(GenericMarker::deleteMarker);

        //Updates all quest area markers
        if (questsChanged) {
            //Update all area markers for quests
            updateQuestAreas();

            //Generate information about all NPCs involved in quests
            generateQuestNPCInfo();

            //Generate markers based on the generated info
            generateAllMarkers();
        }
    }

    /**
     * Generates information about all NPCs involved in quests
     */
    private void generateQuestNPCInfo() {
        //Clear any previous information
        questGiverInfo = new HashMap<>();
        //Generation information about NPC's parts in each quest
        for (IQuest quest : questsAPI.getLoadedQuests()) {
            if (quest.getNpcStart() != null) {
                getInfo(quest.getNpcStart()).addQuestStart(quest);
            }
            for (IStage stage : quest.getStages()) {
                for (UUID npcId : stage.getNpcsToKill()) {
                    getInfo(npcId).addQuestKill(quest);
                }
                for (UUID npcId : stage.getItemDeliveryTargets()) {
                    getInfo(npcId).addQuestDeliver(quest);
                }
                for (UUID npcId : stage.getNpcsToInteract()) {
                    getInfo(npcId).addQuestInteract(quest);
                }
            }
        }
    }

    /**
     * Generates all quest markers based on the previously generated quest NPC info
     */
    private void generateAllMarkers() {
        NPCRegistry registry = CitizensAPI.getNPCRegistry();
        //Add markers for each NPC detected as part of a quest
        for (UUID npcId : questGiverInfo.keySet()) {
            NPCQuestInfo info = questGiverInfo.get(npcId);
            MarkerIcon icon = markerIcons.get(info.getNPCIcon());
            List<IQuest> questStarts = info.getQuestStarts();
            StringBuilder markerDescription = new StringBuilder();

            markerDescription.append("<h2>").append(registry.getByUniqueId(npcId).getName()).append("</h2>");

            if (!questStarts.isEmpty()) {
                markerDescription.append("<h3>Quests offered:</h3><ul>");
                for (IQuest quest : questStarts) {
                    markerDescription.append("<li><h4><b>").append(quest.getName()).append("</b></h4><h5><b>- ");
                    markerDescription.append(quest.getDescription()).append("</b></h5>").append(
                            new QuestStagesInfoGenerator(quest).getQuestStagesInfo());
                    markerDescription.append(new QuestRewardsInfoGenerator(quest).getQuestRewardsInfo());
                    markerDescription.append(new QuestRequirementsInfoGenerator(quest).getQuestRequirementsInfo());
                    markerDescription.append(new QuestPlannerInfoGenerator(quest).getQuestPlannerInfo()).append("</li>");
                }
                markerDescription.append("</ul>");
            }

            markerDescription.append(getInvolvedInQuestsString(info));

            addNPCMarker(npcId, QuestsHelper.getMarkerTitle(info.getQuestNPCType()), markerDescription.toString(), icon, questMarkerSet);
        }
    }

    /**
     * Gets information about the quests an NPC is involved in
     *
     * @param info <p>The NPC info to look through</p>
     * @return <p>Information about an NPC's involvement in different quests</p>
     */
    private String getInvolvedInQuestsString(NPCQuestInfo info) {
        List<IQuest> questKills = info.getQuestKills();
        List<IQuest> questInteractions = info.getQuestInteractions();
        List<IQuest> questDeliveries = info.getQuestDeliveries();

        if (questKills.isEmpty() && questInteractions.isEmpty() && questDeliveries.isEmpty()) {
            return "";
        }
        StringBuilder markerDescription = new StringBuilder();
        markerDescription.append("<h3>Involved in quests:</h3><ul>");
        addInvolvedInString("Killed in", questKills, markerDescription);
        addInvolvedInString("Delivery target in", questDeliveries, markerDescription);
        addInvolvedInString("Interacted with in quest", questInteractions, markerDescription);
        markerDescription.append("</ul>");
        return markerDescription.toString();
    }

    /**
     * Adds string explaining how an NPC is involved with the given quests
     *
     * @param prefix  <p>The string explaining how the NPC is involved</p>
     * @param quests  <p>The quests the NPC is involved in</p>
     * @param builder <p>The string builder to append to</p>
     */
    private void addInvolvedInString(String prefix, List<IQuest> quests, StringBuilder builder) {
        for (IQuest quest : new HashSet<>(quests)) {
            builder.append("<li>").append(prefix).append(": ").append(quest.getName()).append("</li>");
        }
    }


    /**
     * Gets the info object for the given NPC
     *
     * @param npcId <p>The id of the NPC to get information about</p>
     * @return <p>The NPC's info object</p>
     */
    private NPCQuestInfo getInfo(UUID npcId) {
        if (questGiverInfo.get(npcId) == null) {
            questGiverInfo.put(npcId, new NPCQuestInfo());
        }
        return questGiverInfo.get(npcId);
    }

    /**
     * Updates all quest area markers
     */
    private void updateQuestAreas() {
        questAreaMarkerSet.getCircleMarkers().forEach(GenericMarker::deleteMarker);
        for (IQuest quest : questsAPI.getLoadedQuests()) {
            for (IStage stage : quest.getStages()) {
                markKillLocations(quest, stage);
                markReachLocations(quest, stage);
            }
        }
        //TODO: Mark WorldGuard areas part of quests. Requires WorldGuard integration
    }

    /**
     * Marks any reach locations found in the given stage
     *
     * @param quest <p>The quest the stage belongs to</p>
     * @param stage <p>The stage to search for reach locations</p>
     */
    private void markReachLocations(IQuest quest, IStage stage) {
        markLocations(stage.getLocationsToReach(), stage.getRadiiToReachWithin(),
                "<b>Target location for:</b> " + quest.getName());
    }

    /**
     * Marks any kill locations found in the given stage
     *
     * @param quest <p>The quest the stage belongs to</p>
     * @param stage <p>The stage to search for kill locations</p>
     */
    private void markKillLocations(IQuest quest, IStage stage) {
        markLocations(stage.getLocationsToKillWithin(), stage.getRadiiToKillWithin(),
                "<b>Kill location for:</b> " + quest.getName());
    }

    /**
     * Marks the given locations on the dynamic map
     *
     * @param locations   <p>The locations to mark</p>
     * @param radii       <p>The radius of each location's circle</p>
     * @param description <p>The description for what the location means</p>
     */
    private void markLocations(List<Location> locations, List<Integer> radii, String description) {
        for (int i = 0; i < locations.size(); i++) {
            Location location = locations.get(i);
            int radius = radii.get(i);

            //Skip if location is invalid
            World world = location.getWorld();
            if (world == null) {
                continue;
            }

            CircleMarker circleMarker = questAreaMarkerSet.createCircleMarker(null, description, true,
                    world.getName(), location.getX(), location.getY(), location.getZ(), radius, radius, false);
            if (circleMarker == null) {
                DynmapCitizens.getInstance().getLogger().log(Level.WARNING, "Unable to create circle marker at " +
                        location + " with radius " + radius);
            } else {
                circleMarker.setFillStyle(0.3, 0x75AFD2);
                circleMarker.setLineStyle(1, 1.0, 0x36c90e);
            }
        }
    }

}
